
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"


#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"


#include "initng_chroot.h"
#include <assert.h>

s_entry CHROOT = { "chroot", STRING, 6 };

int do_chroot(active_h * s, process_h * p __attribute__ ((unused)))
{
    char *tmp = NULL;

    assert(s);
    assert(p);

    D_("do_suid!\n");
    if (!(tmp = active_db_get_string(&CHROOT, s)))
    {
        D_("SUID not set!\n");
        return (TRUE);
    }

    if (chdir(tmp) == -1)
    {
        F_("Chdir failed with %s\n", strerror(errno));
        return (FALSE);
    }
    if (chroot(tmp) == -1)
    {
        F_("Chroot failed with %s\n", strerror(errno));
        return (FALSE);
    }
    return (TRUE);
}

#ifdef STATIC_CHROOT
int chroot_module_init(const char *version)
{
#else
int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

#endif
    initng_sdt_add(&CHROOT);
    return (initng_add_hook(A_FORK, 50, &do_chroot));
}

#ifdef STATIC_CHROOT
void chroot_module_unload(void)
#else
void module_unload(void)
#endif
{
    D_("module_unload();\n");
    initng_sdt_del(&CHROOT);
    initng_del_hook(A_FORK, &do_chroot);
}
