/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"

#include "initng_depend.h"

s_entry ALSO_START = { "also_start", STRINGS, 10 };
s_entry ALSO_STOP = { "also_stop", STRINGS, 9 };


/* returns TRUE if all use deps are started */
static int check_use(active_h * service)
{
    active_h *tmp_service = NULL;
    char *use = NULL;
    char *need = NULL;

    /* Handle NEED */
    while ((need = active_db_get_next_string(&NEED, service, need)))
    {
        D_("NEED : setting NEED %s in %s\n", need, service->name);
        if (!(tmp_service = active_db_find_by_name(need)))
        {
            /* why are we wanna load modules, from disk again?, if they did not exits
               when this service was started, it wont exist on disk now either, its just
               waisting cpu cycles */
#ifdef TRY_LOAD_MODULES_AGAIN
            /* hack, check if it contains virtual/ beginning, if we wont find it anyway */
            /* it will work w-o this hack, but it will print, that needed service is not found,
               and it will delay, looking for a service we now wont exist */
            if (strncmp(need, "virtual/", 8) == 0)
                return (FALSE);

            if ((tmp_service = load_new_service_named(need)))   /* needed service may not be loaded yet */
                start_service(tmp_service);
            else
#endif
            {
                /* this might be ok, if we are waiting for an virtual, or to be started. */
                D_("Cound not load needed dependecy %s, depended by %s !\n",
                   need, service->name);
                /* runlevel shud NEVER fail, on missing service */
                /* if(strcmp(service->name, g.runlevel)==0)
                   continue; */
                return (FALSE);
            }
        }

        switch (tmp_service->a_status)
        {
            case RUNNING:
            case DONE:
                continue;                   /* this use is met */
            default:
                D_("\"%s\" still waits for use \"%s\" that has status \"%s\".\n", service->name, tmp_service->name, active_db_get_status_string(tmp_service->a_status));
                return (FALSE);
        }
    }

    /* Handle USE */
    while ((use = active_db_get_next_string(&USE, service, use)))
    {
        D_("USE : setting USE %s in %s\n", use, service->name);
        tmp_service = active_db_find_by_name(use);
        if (!tmp_service)
            continue;
        switch (tmp_service->a_status)
        {
            case RUNNING:
            case DONE:
                continue;                   /* this use is met */
            default:
                D_("\"%s\" still waits for use \"%s\" that has status \"%s\".\n", service->name, tmp_service->name, active_db_get_status_string(tmp_service->a_status));
                return (FALSE);
        }
    }

    return (TRUE);
}

int service_state(active_h * service)
{
    char *tmp = NULL;
    active_h *current;

    switch (service->a_status)
    {
        case LOADING:
            {
                char *cs = NULL;

                D_("service %s loading\n", service->name);
                while ((cs = active_db_get_next_string(&NEED, service, cs)))
                {
                    /* hack, check if it contains virtual/ beginning, if we wont find it anyway */
                    /* it will work w-o this hack, but it will print, that needed service is not found,
                       and it will delay, looking for a service we now wont exist */
                    if (strncmp(cs, "virtual/", 8) == 0)
                        continue;
                    D_("needed: %s\n", cs);
                    if (!start_new_service_named(cs))
                    {
                        F_("%s need dep %s coud not start!\n", service->name,
                           cs);

#ifdef FAIL_IF_DEP_CANT_LOAD
                        /* if this is runlevel, never complain here */
                        if (strcmp(service->name, g.runlevel) == 0)
                        {
                            continue;
                        }

                        mark_service(service, START_DEP_FAILED);
                        active_db_del(service);
                        active_db_free(service);
                        return (FALSE);
#else
                        continue;
#endif
                    }
                }

            }
            break;
        case START_DEP_MET:
            if (check_use(service) == FALSE)
            {
                mark_service(service, WAITING_FOR_START_DEP);
            }
            break;
        case STARTING:
            /* Handle ALSO_START  (these are not dependencies) */
            while ((tmp =
                    active_db_get_next_string(&ALSO_START, service, tmp)))
                start_new_service_named(tmp);
            /* todo, by listening to STOP_MARKED we can try to stop stuff a bit earlier? */
            break;
        case STOPPED:
            /* Handle ALSO_STOP */
            while ((tmp =
                    active_db_get_next_string(&ALSO_STOP, service, tmp)))
                if ((current = active_db_find_by_name(tmp)))
                {
                    D_("service %s also stops %s\n", service->name, tmp);
                    stop_service(current);
                }
            break;
        default:
            break;
    }
    return (TRUE);
}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_sdt_add(&ALSO_START);
    initng_sdt_add(&ALSO_STOP);
    initng_add_hook(WATCHERS, 50, &service_state);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_sdt_del(&ALSO_START);
    initng_sdt_del(&ALSO_STOP);
    initng_del_hook(WATCHERS, &service_state);
}
