
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() pid_t */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>
#include <errno.h>


#include "initng_active_db.h"
#include "initng_process_db.h"
#include "initng_signal.h"
#include "initng_plugin_callers.h"
#include "initng_execute.h"

#include "initng_fork.h"



pid_t initng_fork(active_h * service, process_h * process)
{
    /* This is the real service kicker */
    pid_t pid_fork;             /* pid got from fork() */
    int try_count = 0;          /* Count tryings */

    assert(service);
    assert(process);

    /* Try to fork 30 times */
    while ((pid_fork = fork()) == -1)
    {
        if (try_count > 30)
        {                                   /* Already tried 30 times = no more try */
            F_("GIVING UP TRY TO FORK, FATAL for service.\n");
            return (pid_fork);
        }
        F_("FAILED TO FORK! try no# %i, this can be FATAL!\n", try_count);
        usleep(++try_count * 2000);         /* Sleep a while before trying again */
    }

    /* if this is child */
    if (pid_fork == 0)
    {

        disable_signals();

#ifdef DEBUG
        /* unverbose by default for now */
        g.verbose = 0;
#endif

        /* do a minimum sleep, to let the mother process 
           to register child, and notice death */
        usleep(ALL_USLEEP);
	

        if (!initng_plugin_after_fork(service, process))
            _exit(1);

        if (g.i_am_init)
        {
            ioctl(0, TIOCNOTTY, 0);
            setsid();                       /* Run a program in a new session ??? */
        }

	/* this will redirect STDOUT and STDERR to where it belongs */
        set_pipes(process);

	/* close all open fds, exept STDIN, STDOUT, STDERR (0, 1, 2) */
	{
	    int i;
	    for(i=3; i<=1013; i++)	
		close(i);
	}


        if (g.i_am_init)
            tcsetpgrp(0, getpgrp());        /* run this in foreground on fd 0 */

    }

    return (pid_fork);
}
