
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <termios.h>
#include <stdio.h>
#include <sys/klog.h>

#include "initng_global.h"
#include "initng_signal.h"
#include "initng_history.h"
#include "initng_handler.h"
#include "initng_execute.h"                 /* new_environ() */
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_toolbox.h"


#include "initng_main.h"
/*extern char **environ; */

/*
 * Poll if there are any processes left, and if not quit initng
 */
int ready_to_quit(void)
{
    D_("ready_to_quit(): %i process(es) remaining.\n", active_db_count(0));

    /* If the last process has died, quit initng */
    if ((active_db_count(NONE)) < 1)
    {
        active_h *tmp = NULL;

        while_active_db(tmp)
        {
            printf
                ("********  %s still active (status: %s) will be ignored *******\n",
                 tmp->name, active_db_get_status_string(tmp->a_status));
            sleep(2);
        }

        return (TRUE);
    }
    return (FALSE);
}

/* this is called when there is no processses left */
void when_out(void)
{
    if (g.i_am_init && getpid() != 1)
    {
        F_("I AM NOT INIT, THIS CANT BE HAPPENING!\n");
	sleep(3);
        return;
    }

    switch (g.when_out)
    {
        case THEN_QUIT:
            printf(" ** Now Quiting **");
            exit_initng(0);
            break;
        case THEN_SULOGIN:
            printf(" ** Now SuLogin");
            su_login();
            break;
        case THEN_RESTART:
            printf(" ** Now restarting");
            restart_initng();
            break;
        case THEN_NEW_INIT:
            printf(" ** Launtching new init");
            new_init();
            break;
        default:
            F_("Error: invalid command (code %d)\n", g.when_out);
            /* FALLTHROUGH */
    }

}

/*
 *  Launch sulogin and wait it for finish
 */
void su_login(void)
{
    pid_t sulogin_pid;
    int status;

    sulogin_pid = fork();

    if (sulogin_pid == 0)
    {
	char *argv[] = { i_strdup("sulogin"), NULL };

	execve("/sbin/sulogin", argv, new_environ(NULL));   /* launch sulogin */
    }

    if (sulogin_pid > 0)
    {
	do {
	    sulogin_pid = waitpid(sulogin_pid, &status, WUNTRACED);
	} while (!WIFEXITED(status) && !WIFSIGNALED(status));
    } else 
	_exit(1); /* sulogin unsuccessfull :( */
}

/*
 * Start initial services
 */
void start_extra_services(void)
{
    int i;
    int a_count = 0;            /* counts orders from argv to start */

    set_sys_state(STATE_STARTING);
    /* check with argv which service to start initiating */
    for (i = 1; i < g.Argc; i++)
    {
        /* start all services that is +service */
        if (g.Argv[i][0] != '+')
            continue;

        /* if succeed to load this service */
        if (start_new_service_named(g.Argv[i] + 1))
            a_count++;
    }
}

void new_init(void)
{
    int i;

    for (i = 3; i <= 1013; i++)
    {
        close(i);
    }
    if (!g.new_init || !g.new_init[0])
    {
        F_(" g.new_init is not set!");
        return;
    }
    fprintf(stderr,
            "\n\n\n          Launching new init (%s)\n\n", g.new_init[0]);
    execve(g.new_init[0], g.new_init, environ);
}

void restart_initng(void)
{
    char **argv = NULL;
    int i;

    for (i = 3; i <= 1013; i++)
    {
        close(i);
    }
    argv = (char **) initng_calloc(3, sizeof(char *));
    set_sys_state(STATE_RESTART);

    argv[0] = (char *) initng_calloc(strlen(g.runlevel) + 12, sizeof(char));
    strcpy(argv[0], "runlevel=");
    strcat(argv[0], g.runlevel);
    argv[1] = NULL;
    fprintf(stderr,
            "\n\n\n          R E S T A R T I N G,  (Really hot reboot)\n\n");
    execve("/sbin/initng", argv, environ);
}

void sys_state_services_loaded(void)
{
    h_sys_state oldstate = g.sys_state;

    set_sys_state(STATE_SERVICES_LOADED);
    g.sys_state = oldstate;
    return;
}


/* this function sets g.sys_state, and call plugins that listen on its change */
void set_sys_state(h_sys_state state)
{
    /* don't set a state that is */
    if (state == g.sys_state)
        return;

    D_("set_sys_state(): %% Setting state to: %i %% \n", state);
    g.sys_state = state;

    /*
     * execute all functions in modules that want to
     * be executed when system state change occurs.
     */

    initng_load_module_system_changed(state);
    return;
}


void exit_initng(int i)
{
    D_("exit_initng();\n");

    set_sys_state(STATE_EXIT);

    initng_unload_all_modules();
    initng_free();
    exit(i);
}
