
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <termios.h>
#include <stdio.h>
#include <sys/klog.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fnmatch.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "initng_global.h"
#include "main.h"
#include "initng_signal.h"
#include "initng_history.h"
#include "initng_handler.h"
#include "initng_execute.h"                 /* new_environ() */
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_service_data_types.h"
#include "initng_fd.h"

/*
 * %%%%%%%%%%%%%%%%%%%%   main ()   %%%%%%%%%%%%%%%%%%%%
 */
#ifdef BUSYBOX
int init_main(int argc, char *argv[], char *env[])
{
#else
int main(int argc, char *argv[], char *env[])
{
#endif
    time_t last;                /* save the time here */
    int hot_reload = FALSE;

#ifdef DEBUG
    int loop_counter = 0;       /* a common used for counter */
#endif

    S_;
    /* set console loglevel */
    klogctl(8, NULL, 1);


    /* enable generation of core files */
    {
        struct rlimit c = { 1000000, 100000 };
        setrlimit(RLIMIT_CORE, &c);
    }

    /* get the time */
    last = time(NULL);

    /* initialise global variables */
    initng_new(argc, argv, env);

    if (getpid() == 1)
        g.i_am_init = TRUE;

    D_("MAIN_PARSE_ARGV\n");
    {
        int i;

        for (i = 0; argv[i]; i++)
        {
            /* set opt to argv */
            char *opt = argv[i];

            /* dont parse options starting with an '+' */
            if (opt[0] == '+')
                continue;
            /*
             * if arg is --verbose, skip the two -- here, and start checking.
             * if arg is -verbose, skip and continue for loop.
             * if arg is verbose, check it below
             */

            /* if its a double --, its ok */
            if (opt[0] == '-' && opt[1] == '-')
                opt += 2;
            /* but a single is not ! */
            if (opt[0] == '-')
                continue;

#ifdef DEBUG
            if (strcmp(opt, "verbose") == 0)
                g.verbose = TRUE;
#endif
            if (strcmp(opt, "i_am_init") == 0)
                g.i_am_init = TRUE;
            if (strcmp(opt, "hot_reload") == 0)
            {
                W_(" Will start after a hot reload ...\n");
                hot_reload = TRUE;
            }
            if (strcmp(opt, "i_am_not_init") == 0)
                g.i_am_init = FALSE;
#define R_L "runlevel="
            if (strncmp(opt, R_L, strlen(R_L)) == 0)
                g.runlevel = i_strdup((opt) + strlen(R_L));
#ifdef DEBUG
#define V_T "verbose_add="
            if (strncmp(opt, V_T, strlen(V_T)) == 0)
                verbose_add((opt) + strlen(V_T));
#endif
        }
    }

    if (g.i_am_init)
    {
        g.when_out = THEN_SULOGIN;
    }
    else
    {
        g.when_out = THEN_QUIT;
    }

    /* See if we are the first process born in the system */
    if (g.i_am_init && !hot_reload)
    {
        int fd;                 /* /dev/console */

        D_("MAIN_SET_I_AM_INIT_STUFF\n");

        reboot(RB_DISABLE_CAD);             /* Disable Ctrl + Alt + Delete */

        /* Try to open the console, but don't control it */
        if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) > 0)
        {
            D_("Opened /dev/console. Setting terminal options.\n");
            ioctl(fd, KDSIGACCEPT, SIGWINCH);   /* Accept signals from 'kbd' */
            close(fd);                      /* Like Ctrl + Alt + Delete signal? */
        }
        else
        {
            D_("Failed to open /dev/console. Setting options anyway.\n");
            ioctl(0, KDSIGACCEPT, SIGWINCH);    /* Accept signals from 'kbd' */
        }

        /* TODO: this block may be incorrect or incomplete */
        /* Q: What does this block really do? */
        {
            struct termios tty;
            int fd;

            /*
             * TODO: /dev/console may still be open from before. Also, if it
             * fails to open, why try to finish the block?
             */
            if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) < 0)
            {
                F_("main(): can't open /dev/console.\n");
            }

            (void) tcgetattr(fd, &tty);

            tty.c_cflag &= CBAUD | CBAUDEX | CSIZE | CSTOPB | PARENB | PARODD;
            tty.c_cflag |= HUPCL | CLOCAL | CREAD;

            tty.c_cc[VINTR] = 3;            /* ctrl('c') */
            tty.c_cc[VQUIT] = 28;           /* ctrl('\\') */
            tty.c_cc[VERASE] = 127;
            tty.c_cc[VKILL] = 24;           /* ctrl('x') */
            tty.c_cc[VEOF] = 4;             /* ctrl('d') */
            tty.c_cc[VTIME] = 0;
            tty.c_cc[VMIN] = 1;
            tty.c_cc[VSTART] = 17;          /* ctrl('q') */
            tty.c_cc[VSTOP] = 19;           /* ctrl('s') */
            tty.c_cc[VSUSP] = 26;           /* ctrl('z') */

            /*
             *      Set pre and post processing
             */
            tty.c_iflag = IGNPAR | ICRNL | IXON | IXANY;
            tty.c_oflag = OPOST | ONLCR;
            tty.c_lflag = ISIG | ICANON | ECHO | ECHOCTL | ECHOPRT | ECHOKE;

            /*
             *      Now set the terminal line.
             *      We don't care about non-transmitted output data
             *      and non-read input data.
             */
            (void) tcsetattr(fd, TCSANOW, &tty);
            (void) tcflush(fd, TCIOFLUSH);
            (void) close(fd);
        }

    }

    D_("MAIN_LOAD_MODULES\n");
    /* Load modules, if fails - launch sulogin and then try again */
    while (!initng_load_all_modules())
    {
        /* if we can't load modules, revert to single-user login */
        su_login();
    }

    /* Configure signal handlers */
    (void) enable_signals();

    /* change system state */
    set_sys_state(STATE_STARTING);

    D_("MAIN_START_DEFAULT_SERVICE\n");
    /* If no starting services found, use default */
    if (!g.runlevel)
    {
        if (g.i_am_init)
        {
            g.runlevel = i_strdup("default");
        }
        else
        {
            g.runlevel = i_strdup("fake-default");
        }
    }

    /* set title of initng, can be watched in ps -ax */
    set_proc_title("initng [%s]", g.runlevel);


    /* make sure this is not a hot reload */
    if (!hot_reload)
    {
        /* first start all services, prompted at boot with +daemon/test */
        start_extra_services();

        /* try load the default service, if it fails - launch sulogin and try again */
        while (!start_new_service_named(g.runlevel))
        {
            F_("Failed to load runlevel (%s)!\n", g.runlevel);
            su_login();
        }
    }

    /* set system state, plugins can hook here */
    sys_state_services_loaded();

    D_("MAIN_GOING_MAIN_LOOP\n");
    /* %%%%%%%%%%%%%%%   MAIN LOOP   %%%%%%%%%%%%%%% */
    for (;;)
    {
        time_t now;             /* Store the start time of each loop */

        D_("MAIN_LOOP: %i\n", loop_counter++);

        /* Update current time */
        now = time(NULL);

        /* call hooks that want to be run every main loop */
        if (initng_plugin_call(MAIN, NULL) < 0)
            F_("Some plugin failed!\n");

        /* Check for sigchld */
        while (initng_signal_got_sigchld)
        {
            D_("There are dead childs to handle!\n");
            initng_signal_got_sigchld = FALSE;
            handle_sigchild();
        }

        /* Check for sig alarm */
        while (initng_signal_got_alarm == TRUE)
        {
            D_("Got alarm set, checking timers plugin ...\n");
            initng_signal_got_alarm = FALSE;
            /* Check for timed out stopping services for SIGKILL */
            term_stopping_services();
            /* Check timers plugins */
            initng_plugin_timers();
        }

        /* If we got an interrupt load ctrlaltdel */
        while (initng_signal_got_ctrl_alt_del)
        {
            D_("I got ctrl+alt+del!\n");
            initng_signal_got_ctrl_alt_del = FALSE;
            if (g.i_am_init)
                start_new_service_named("ctrlaltdel");
            else
                start_new_service_named("ctrlaltdel-fake");
        }

        /* If we got a keyboard request load kbreq */
        while (initng_signal_got_keyboard_request)
        {
            D_("I got a keyboard request!\n");
            initng_signal_got_keyboard_request = FALSE;
            start_new_service_named("kbreq");
        }

        /*
         * If now is before last, the time was set backwards.
         * If so, compensate the change in the whole process_db
         */
        if (now < last)
        {
            D_(" Now is before last time, clock was screwed backwards. Compensating for %i seconds.\n", last - now);
            active_db_compensate_time(last - now);
            history_db_compensate_time(last - now);
        }

        /*
         * If we moved more than an hour ahead, work backwards in one hour
         * chunks until we are within one hour of the original.
         *
         * TODO: why is this important?
         */
        while (now - last >= 3600)
        {
            D_(" It has gone i hour sence last loop, this cant be true. Compensating for 3600 seconds.\n");
            active_db_compensate_time(3600);
            history_db_compensate_time(3600);
            last = last + 3600;
        }
        last = now;                         /* Remember last time */

        /* 
         * Start launching dependencies, until no more services are marked.
         * The dependencies db should continue to stabilize through each run.
         */
        while (g.interrupt)
        {
            D_("Resetting g.interrupt = FALSE\n");
            g.interrupt = FALSE;
            try_start_wait_deps();
            D_("mini loop (will continue as long as interrupt is set)\n");
        }
        D_("Finished loading dependencies!\n");
        if (initng_signal_got_sigchld || initng_signal_got_keyboard_request ||
            initng_signal_got_alarm || initng_signal_got_ctrl_alt_del)
            g.interrupt = TRUE;

        /*
           while (g.modules_to_unload==TRUE)
           {
           g.modules_to_unload=FALSE;
           D_("There is modules to unload!\n");
           initng_unload_marked_modules();
           }
         */



        /*
         * Check if there are any running processes left, otherwise quit, when_out try to do something when this happends.
         */

        if (ready_to_quit() == TRUE)
        {
            set_sys_state(STATE_ASE);
            printf(" *** All services gone. ***\n");
            when_out();
            continue;
        }

        /*
         * See if anything is happening on the control channel
         *  - wait for 60seconds
         */

        /* if interupt is set, do a quick poll */
        if (g.interrupt)
        {
            initng_plugin_fd_poll(0);
        }
        else
        {
            /* if system is shuting down, do faster polls */
            if (g.sys_state == STATE_STOPPING)
            {
                if (!initng_plugin_fd_poll(20))
                {
                    F_("Waited for 3 minutes when shutting down, setting verbose mode!\n");
#ifdef DEBUG
                    g.verbose = TRUE;
#endif
                }
            }
            /* else, wait forever - poll will be aborted by signals anyway. */
            else
            {
                initng_plugin_fd_poll(60000);
            }
        }
    }                                       /* End main loop */
}
